﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Web;

using App6.Dao;
using System.Reflection;
using NHibernate.Criterion;

using NHibernate;
using Newtonsoft.Json;

namespace App6.Utils
{
    public class DataSource
    {
        public string ID { get; set; }
        public string dataFormat { get; set; }
        public ICollection<Dictionary<string, object>> fields { get; set; }
        public ICollection<Dictionary<string, object>> operationBindings { get; set; }
        public string beanClassName { get; set; }

        /// <summary>
        /// Get a field's data by it's name
        /// </summary>
        /// <param name="name">The name of the field</param>
        /// <returns>A Dictionary with the attributes of the field, or null if field was not found</returns>
        public Dictionary<string, object> getField(string name)
        {
            foreach (Dictionary<string, object> field in fields)
            {
                if (field.ContainsKey("name") && field["name"].Equals(name))
                {
                    return field;
                }
            }

            return null;
        }

        public string getPrimaryKey()
        {
            foreach (Dictionary<string, object> field in fields)
            {
                if (field.ContainsKey("primaryKey") && field["primaryKey"].Equals("true"))
                {
                    return field["name"] as string;
                }
            }

            return null;
        }

        virtual public DSResponse execute<T>(DSRequest<T> request)
        {
            if (request == null)
            {
                return null;
            }

            if ("add".Equals(request.operationType))
            {
                return executeAdd<T>(request);
            }

            if ("fetch".Equals(request.operationType))
            {
                return executeFetch<T>(request);
            }

            if ("update".Equals(request.operationType))
            {
                return executeUpdate<T>(request);
            }

            if ("remove".Equals(request.operationType))
            {
                return executeRemove<T>(request);
            }

            return null;
        }

        virtual public DSResponse executeAdd<T>(DSRequest<T> request)
        {
            Type beanType = Type.GetType(beanClassName);

            if (beanType == null)
            {
                return null;
            }

            var req = ReflectionHelper.InvokeGeneric(request.RpcManager, "convertRequest", beanType, request);
            var data = ReflectionHelper.GetProperty(req, "data");

            request.RpcManager.Session.Save(ReflectionHelper.GetProperty(req, "data"));
            request.RpcManager.Session.Flush();

            DSResponse dsresponse = new DSResponse();

            dsresponse.data = data;
            dsresponse.status = 0;

            return dsresponse;
        }

        virtual public DSResponse executeRemove<T>(DSRequest<T> request)
        {
            Type beanType = Type.GetType(beanClassName);

            if (beanType == null)
            {
                return null;
            }

            var req = ReflectionHelper.InvokeGeneric(request.RpcManager, "convertRequest", beanType, request);
            var data = ReflectionHelper.GetProperty(req, "data");

            var itm = request.RpcManager.Session.Get(beanType,ReflectionHelper.GetProperty(data, getPrimaryKey()));

            request.RpcManager.Session.Delete(itm);
            request.RpcManager.Session.Flush();

            DSResponse dsresponse = new DSResponse();

            dsresponse.data = data;
            dsresponse.status = 0;

            return dsresponse;
        }

        virtual public DSResponse executeUpdate<T>(DSRequest<T> request)
        {
            DSRequest<Dictionary<string, object>> reqmap = request as DSRequest<Dictionary<string, object>>;
            Type beanType = Type.GetType(beanClassName);

            if (beanType == null)
            {
                return null;
            }

            var req = ReflectionHelper.InvokeGeneric(request.RpcManager, "convertRequest", beanType, request);
            var reqdata = ReflectionHelper.GetProperty(req, "data");

            var itm = request.RpcManager.Session.Get(beanType, ReflectionHelper.GetProperty(reqdata, getPrimaryKey()));

            // update all fields which have changed. they are defined in the data property
            // build the criteria
            if (reqmap.data.Keys.Count != 0)
            {
                foreach (string key in reqmap.data.Keys)
                {
                    if (getField(key) != null)
                    {
                        object val = ReflectionHelper.GetProperty(reqdata, key);
                        ReflectionHelper.SetProperty(itm, key, val);
                    }
                }
            }

            request.RpcManager.Session.Update(itm);
            request.RpcManager.Session.Flush();

            // create a result object to be returned
            // and copy all properties of the updated object into this one
            var data = ReflectionHelper.CreateInstance(beanClassName);

            if (reqmap.oldValues.Keys.Count != 0)
            {
                foreach (string key in reqmap.oldValues.Keys)
                {
                    if (getField(key) != null)
                    {
                        object val = ReflectionHelper.GetProperty(itm, key);
                        ReflectionHelper.SetProperty(data, key, val);
                    }
                }
            }

            // create the DSResponse object
            DSResponse dsresponse = new DSResponse();
            dsresponse.data = data;
            dsresponse.status = 0;

            return dsresponse;
        }

        virtual public DSResponse executeFetch<T>(DSRequest<T> request)
        {
            
            DSRequest<Dictionary<string, object>> req = request as DSRequest<Dictionary<string, object>>;

            Type beanType = Type.GetType(beanClassName);

            if (beanType == null)
            {
                return null;
            }

            var itmreq = ReflectionHelper.InvokeGeneric(request.RpcManager, "convertRequest", beanType, req);

            var query = request.RpcManager.Session.CreateCriteria(beanType);

            if (req.advancedCriteria == null)
            {
                buildStandardCriteria(itmreq, req, ref query);
            }
            else
            {
                buildAdvancedCriteria(req, ref query);
            }

            // build the criteria
            if (req.data.Keys.Count != 0)
            {
                foreach (string key in req.data.Keys)
                {
                    Dictionary<string, object> field = getField(key);

                    // make sure the field is in the DataSource
                    if (field != null)
                    {
                        object val = ReflectionHelper.GetProperty(itmreq, key);

                        string type = field["type"] as string;

                        if (type.Equals("text") || type.Equals("link") || type.Equals("enum") ||
                            type.Equals("image") || type.Equals("ntext"))
                        {
                            query.Add(Restrictions.Like(key, "%" + val + "%"));
                            break;
                        }
                        else
                        {
                            query.Add(Restrictions.Eq(key, val));
                            break;
                        }
                    }
                }
            }

            // add sorting
            if (req.sortBy != null)
            {
                // add the sorting
                foreach (string column in req.sortBy)
                {
                    // if column name is with -, then ordering is descending, otherwise ascending 
                    if (column.StartsWith("-"))
                    {
                        // if sort is descending, then we have to remove the '-' from the beginning to get te correct
                        // column name
                        query.AddOrder(Order.Desc(column.Substring(1)));
                    }
                    else
                    {
                        query.AddOrder(Order.Asc(column));
                    }
                }
            }

            // create a response object
            DSResponse dsresponse = new DSResponse();

            // set start row and number of rows on the query itself
            if (req.endRow != 0)
            {
                query.SetMaxResults(req.endRow - req.startRow);
                query.SetFirstResult(req.startRow);
            }

            // get the requested number of objects
            var products = query.List();

            // change projection to get the total number of rows
            // we need to clear the result range and the ordering for this to work
            query.SetProjection(Projections.RowCount());
            query.SetFirstResult(0);
            query.SetMaxResults(int.MaxValue);
            query.ClearOrders();

            // set total rows using the projection
            dsresponse.totalRows = (int)query.UniqueResult<int>();

            // set the response data
            dsresponse.data = products;
            dsresponse.startRow = req.startRow;
            dsresponse.endRow = dsresponse.startRow + (ReflectionHelper.GetProperty(products, "Count") as int?).Value ;

            // sanity check, if no rows, return 0
            if (dsresponse.endRow < 0)
            {
                dsresponse.endRow = 0;
            }

            dsresponse.status = 0;

            return dsresponse;
        }

        void buildAdvancedCriteria(DSRequest<Dictionary<string, object>> req, ref ICriteria query)
        {
            if (req.advancedCriteria != null)
            {
                AbstractCriterion restriction = buildCriterion(req.advancedCriteria);
                query.Add(restriction);
            }
        }

        void fixValueType(string fn, ref object val)
        {
            Dictionary<string, object> field = getField(fn);

            if (field.ContainsKey("type"))
            {
                string type = field["type"] as string;

                if ("date".Equals(type))
                {
                    val = JsonConvert.DeserializeObject("\"" + val as string + "\"", typeof(DateTime));
                }

                if (val.GetType() == typeof(long))
                {
                    long? lval = val as long?;
                    if ("float".Equals(type))
                    {
                        val = (float)lval.Value;
                    }
                    if ("double".Equals(type))
                    {
                        val = (double)lval.Value;
                    }
                    if ("int".Equals(type))
                    {
                        val = (int)lval.Value;
                    }
                }
                if (val.GetType() == typeof(int))
                {
                    int? ival = val as int?;
                    if ("float".Equals(type))
                    {
                        val = (float)ival.Value; ;
                    }
                    if ("double".Equals(type))
                    {
                        val = (double)ival.Value; ;
                    }
                }
            }
        }

        /// <summary>
        /// Recursive function to parse an advanced criteria and build up the NHibernate criteria from it
        /// </summary>
        /// <param name="criteria">The advanced criteria from which we need to build the NHibernate criteria object</param>
        /// <returns></returns>
        AbstractCriterion buildCriterion(AdvancedCriteria criteria)
        {
            string fn = criteria.fieldName;
            string op = criteria.Operator;
            object val = criteria.value;

            // fix type of value to the type specified in field declaration, if required
            if (fn != null)
            {
                fixValueType(fn, ref val);
            }

            // case when simple criterion
            if (criteria.criteria == null)
            {
                if ("equals".Equals(op))
                {
                    return Restrictions.Eq(fn, val);
                }
                else if ("notEqual".Equals(op))
                {
                    return Restrictions.Not(Restrictions.Eq(fn, val));
                }
                if ("iEquals".Equals(op))
                {
                    return Restrictions.Eq(fn, val).IgnoreCase();
                }
                else if ("iNotEqual".Equals(op))
                {
                    return Restrictions.Not(Restrictions.Eq(fn, val).IgnoreCase());
                }
                else if ("greaterThan".Equals(op))
                {
                    return Restrictions.Gt(fn, val);
                }
                else if ("lessThan".Equals(op))
                {
                    return Restrictions.Lt(fn, val);
                }
                else if ("greaterOrEqual".Equals(op))
                {
                    return Restrictions.Ge(fn, val);
                }
                else if ("lessOrEqual".Equals(op))
                {
                    return Restrictions.Le(fn, val);
                }
                else if ("contains".Equals(op))
                {
                    return Restrictions.Like(fn, "%" + val + "%");
                }
                else if ("startsWith".Equals(op))
                {
                    return Restrictions.Like(fn, val + "%");
                }
                else if ("endsWith".Equals(op))
                {
                    return Restrictions.Like(fn, "%" + val);
                }
                else if ("iContains".Equals(op))
                {
                    return Restrictions.InsensitiveLike(fn, "%" + val + "%");
                }
                else if ("iStartsWith".Equals(op))
                {
                    return Restrictions.InsensitiveLike(fn, val + "%");
                }
                else if ("iEndsWith".Equals(op))
                {
                    return Restrictions.InsensitiveLike(fn, "%" + val);
                }
                else if ("notContains".Equals(op))
                {
                    return Restrictions.Not(Restrictions.Like(fn, "%" + val + "%"));
                }
                else if ("notStartsWith".Equals(op))
                {
                    return Restrictions.Not(Restrictions.Like(fn, val + "%"));
                }
                else if ("notEndsWith".Equals(op))
                {
                    return Restrictions.Not(Restrictions.Like(fn, "%" + val));
                }
                else if ("iNotContains".Equals(op))
                {
                    return Restrictions.Not(Restrictions.InsensitiveLike(fn, "%" + val + "%"));
                }
                else if ("iNotStartsWith".Equals(op))
                {
                    return Restrictions.Not(Restrictions.InsensitiveLike(fn, val + "%"));
                }
                else if ("iNotEndsWith".Equals(op))
                {
                    return Restrictions.Not(Restrictions.InsensitiveLike(fn, "%" + val));
                }
                else if ("isNull".Equals(op))
                {
                    return Restrictions.IsNull(fn);
                }
                else if ("notNull".Equals(op))
                {
                    return Restrictions.IsNotNull(fn);
                }
                else if ("equalsField".Equals(op))
                {
                    return Restrictions.EqProperty(fn, val as string);
                }
                else if ("notEqualField".Equals(op))
                {
                    return Restrictions.Not(Restrictions.EqProperty(fn, val as string));
                }
                else if ("greaterThanField".Equals(op))
                {
                    return Restrictions.GtProperty(fn, val as string);
                }
                else if ("lessThanField".Equals(op))
                {
                    return Restrictions.LtProperty(fn, val as string);
                }
                else if ("greaterOrEqualField".Equals(op))
                {
                    return Restrictions.GeProperty(fn, val as string);
                }
                else if ("lessOrEqualField".Equals(op))
                {
                    return Restrictions.LeProperty(fn, val as string);
                }
                else if ("between".Equals(op))
                {
                    //greaterThan + lessThan + and
                    return Restrictions.And(
                        Restrictions.Gt(fn, criteria.start),
                        Restrictions.Lt(fn, criteria.end)
                        );
                }
                if (criteria.Operator.Equals("betweenInclusive"))
                {
                    return Restrictions.And(Restrictions.Ge(criteria.fieldName, criteria.start),
                        Restrictions.Le(criteria.fieldName, criteria.end));
                }

                return null;
            }
            else
            {
                // composed criteria
                List<AbstractCriterion> crits = new List<AbstractCriterion>();

                // build the list of subcriterias or criterions
                foreach (AdvancedCriteria crit in criteria.criteria)
                {
                    AbstractCriterion itm = buildCriterion(crit);
                    crits.Add(itm);
                }

                // return this criteria, using the right operator

                if ("and".Equals(criteria.Operator))
                {
                    Conjunction cj = Restrictions.Conjunction();
                    foreach (AbstractCriterion c in crits)
                    {
                        cj.Add(c);
                    }

                    return cj;
                }

                if ("or".Equals(criteria.Operator))
                {
                    Disjunction dj = Restrictions.Disjunction();
                    foreach (AbstractCriterion c in crits)
                    {
                        dj.Add(c);
                    }

                    return dj;
                }

                if ("not".Equals(criteria.Operator))
                {
                    Disjunction dj = Restrictions.Disjunction();
                    foreach (AbstractCriterion c in crits)
                    {
                        dj.Add(Restrictions.Not(c));
                    }

                    return dj;
                }

            }

            return null;
        }

        void buildStandardCriteria(object req, DSRequest<Dictionary<string, object>> reqmap, ref ICriteria query)
        {
            // build the criteria
            if (reqmap.data.Keys.Count != 0)
            {
                foreach (string key in reqmap.data.Keys)
                {
                    Dictionary<string, object> field = getField(key);

                    // make sure the field is in the DataSource
                    if (field != null)
                    {
                        // get the property value
                        object data = ReflectionHelper.GetProperty(req, "data");
                        object val = ReflectionHelper.GetProperty(data, key);

                        string type = field["type"] as string;

                        if (type.Equals("text") || type.Equals("link") || type.Equals("enum") ||
                            type.Equals("image") || type.Equals("ntext"))
                        {
                            query.Add(Restrictions.Like(key, "%" + val + "%"));
                            break;
                        }
                        else
                        {
                            query.Add(Restrictions.Eq(key, val));
                            break;
                        }
                    }
                }
            }
        }
    }
}